@@ -17,6 +17,11 @@ print.jeroen <- function(x, title = paste0("<", is(x), ">"), ...){
   get(y, x, inherits = FALSE)
 }
 
+#' @export
+names.jeroen <- function(x, ...){
+  ls(x, ...)
+}
+
 #' @export
 `[[.jeroen` <- `$.jeroen`
 
